module Day7 (
  buildTree,
  buildTree',
  mkdir,
  Directory (..),
  calculateSize,
  filterDirectories,
  sumUp,
  day7
) where

import Day7.Parser
import Data.Map (Map)
import qualified Data.Map as Map

data Directory = Directory
  { sub :: Map String Directory
  , files :: Map String Int
  , isRoot :: Bool
} deriving (Eq)


instance Show Directory where
  show d = show (files d) ++ " - " ++ show (sub d) ++ "\n"

mkdir :: Directory
mkdir = Directory { sub = Map.empty, files = Map.empty, isRoot = False }

rootDirectory :: Directory
rootDirectory = Directory { sub = Map.empty, files = Map.empty, isRoot = True  }

buildTree :: [TerminalCommand] -> Directory
buildTree commands = fst (buildTree' rootDirectory commands)

buildTree' :: Directory -> [TerminalCommand] -> (Directory, [TerminalCommand])
buildTree'  dir [] = (dir, [])
buildTree' dir (command:cs) =  case command of
  Listing entries ->
    let asFiles = Map.fromList $ toFiles entries
        newDir = dir { files = asFiles }
    in buildTree' newDir cs
  In dirName ->
   let subFolders = sub dir
       selectedDir = Map.findWithDefault mkdir  dirName subFolders
       (replacement, rest) = buildTree' selectedDir cs
       updatedSub = Map.insert dirName replacement subFolders
   in buildTree' dir { sub = updatedSub} rest
  Out -> (dir, cs)
  Root -> if isRoot dir
            then buildTree' dir cs
            else (dir, command:cs)

toFiles :: [ListingEntry] -> [(String, Int)]
toFiles ((FileListing name size):rest) = (name, size): toFiles rest
toFiles (_:rest) = toFiles rest
toFiles [] = []

sizeThreshold :: Int
sizeThreshold = 100000

calculateSize :: Directory -> Int
calculateSize dir = sum (Map.elems (files dir)) + sum sizes
  where
    subFolders = Map.elems (sub dir)
    sizes = map calculateSize subFolders

flatten :: Directory -> [(String, Directory)]
flatten dir = flatten' ("/", dir)
  where
  flatten' (name, d) = (name, d) : concatMap flatten' (Map.toList (sub d))

sizeOfDirectories :: Directory -> [(String, Int)]
sizeOfDirectories dir = map withSize allDirectories
  where
    allDirectories = flatten dir
    withSize (name, directory) = (name, calculateSize directory)

filterDirectories :: (Int -> Bool) -> Directory -> [(String, Int)]
filterDirectories predicate dir = filter (\(_, size) -> predicate size) $ sizeOfDirectories dir

sumUp :: [(String, Int)] -> Int
sumUp = sum . map snd

forceRight :: Either a b -> b
forceRight (Left _) = error "forced Right but got Left"
forceRight (Right b) = b

day7 :: IO ()
day7 = do
   input <- readFile "ressources/day07-input"
   putStrLn "Day7"
   let parsed = forceRight $ parseTerminalLines input
   let tree = buildTree parsed
   let small = filterDirectories (<= 100000) tree
   let sumOfSmall = sumUp small
   putStrLn ("Sum of those directories is " ++ show sumOfSmall)
   let rootSize = snd . head $ sizeOfDirectories tree
   putStrLn ("Root size is " ++ show rootSize)
   let freeSpace = 70000000 - rootSize
   let requiredSpace = 30000000 - freeSpace
   putStrLn ("Required space is " ++ show requiredSpace)
   let bigEnough = filterDirectories (>= requiredSpace) tree
   let onlySizes = map snd bigEnough
   let smallest = minimum onlySizes
   putStrLn ("Smallest directory that is still large enough " ++ show smallest)

